#!/bin/python
# -*- coding: utf-8 -*-
"""
get_data.py
Functions to get data to be displayed in the a pane.

Author: GrimAndGreedy
License: MIT
"""

from listpick.pane.pane_utils import get_file_attributes

def data_refresh_randint_by_row(data, state):
    """
    Add a random number to the data if row id is the same.


    Can be used with right_split_graph()
    
    data[0]: 0,1,2,...,n+1
    data[1]: randint(), randint(), ...
    data[2]: row id
    """
    from random import randint
    if state["indexed_items"]:
        id = state["indexed_items"][state["cursor_pos"]][1][state["id_column"]]
    else:
        return [[0], [0], -1]
    if data in [[], {}, None] or data[2] != id:
        return [[0], [randint(0, 1000)], id]
    else:
        data[0].append(data[0][-1]+1)
        data[1].append(randint(0, 1000))
    return data

def data_refresh_randint(data, state):
    """ 
    Add a random number to the data--regardless of whether the row id is the same.


    Can be used with right_split_graph()

    data[0]: 0,1,2,...,n+1
    data[1]: randint(), randint(), ...
    data[2]: row id
    """
    from random import randint

    if data in [[], {}, None]:
        return [[0], [randint(0, 1000)], -1]
    else:
        data[0].append(data[0][-1]+1)
        data[1].append(randint(0, 1000))
    return data


def data_refresh_randint_title(data, state):
    """ 
    Add a random number to data[1]. 

    Can be used with right_split_display_list()

    data[0]: title
    data[1]: randint(), randint(), ...
    """
    from random import randint

    if data in [[], {}, None]:
        return ["I CHOOSE", [str(randint(0, 1000))]]
    else:
        data[1].append(str(randint(0, 1000)))
    return data

def get_dl(data, state):
    """
    Get dl speed and add it to data[1]

    data[0]: 0,1,2,...,n+1
    data[1]: dl_speed_at_0, dl_speed_at_1, ...
    data[2]: row id
    """
    from aria2tui.utils import aria2c_utils

    gid = "3c60b0fc67230b66"
    req = aria2c_utils.tellStatus(gid)
    info = aria2c_utils.sendReq(req)
    dl = info["result"]["downloadSpeed"]
    ul = info["result"]["uploadSpeed"]

    dl, ul = int(dl), int(ul)

    if data in [[], {}, None]:
        return [[0], [dl], gid]
    else:
        data[0].append(data[0][-1]+1)
        data[1].append(dl)
    return data


def update_file_attributes(data, state):
    """
    Get file attributes

    data[0]: ["size: {}", filetype: {}, last modified: {}]
    ]
    data[1]: id
    """
    if state["indexed_items"]:
        # id = state["indexed_items"][state["cursor_pos"]][1][state["id_column"]]
        id = state["indexed_items"][state["cursor_pos"]][1][0]
    else:
        return [[], -1]
    return [get_file_attributes(id), id]
    




#!/bin/python
# -*- coding: utf-8 -*-
"""
keys.py
Define key dictionary for Picker object.

Author: GrimAndGreedy
License: MIT
"""

import curses
from listpick.utils import keycodes

picker_keys = {
    "refresh":                          [curses.KEY_F5],
    "help":                             [ord('?')],
    "exit":                             [ord('q')],
    "full_exit":                        [3], # Ctrl+c
    # "move_column_left":                 [ord('{')],
    # "move_column_right":                [ord('}')],
    "cursor_down":                      [ord('j'), curses.KEY_DOWN],
    "cursor_up":                        [ord('k'), curses.KEY_UP],
    "half_page_up":                     [ord('u')],
    "half_page_down":                   [ord('d')],
    "page_up":                          [curses.KEY_PPAGE, 2], # Ctrl+b
    "page_down":                        [curses.KEY_NPAGE, 6], # Ctrl+f
    "cursor_bottom":                    [ord('G'), curses.KEY_END],
    "cursor_top":                       [ord('g'), curses.KEY_HOME],
    "five_up":                          [ord('K'), keycodes.META_k],
    "five_down":                        [ord('J'), keycodes.META_j],
    "toggle_select":                    [ord(' ')],
    "select_all":                       [ord('m'), 1], # Ctrl-a
    "select_none":                      [ord('M'), 18],   # Ctrl-r
    "visual_selection_toggle":          [ord('v')],
    "visual_deselection_toggle":        [ord('V')],
    "enter":                            [ord('\n'), curses.KEY_ENTER, 13],
    "redraw_screen":                    [12], # Ctrl-l
    "cycle_sort_method":                [ord('s')],
    "cycle_sort_method_reverse":        [ord('S')],
    "cycle_sort_order":                 [ord('t')],
    "delete":                           [curses.KEY_DC],
    "delete_column":                    [383], # Shift+Delete
    "decrease_lines_per_page":          [ord('-')],
    "increase_lines_per_page":          [ord('+')],
    "increase_column_width":            [ord(']')],
    "decrease_column_width":            [ord('[')],
    "filter_input":                     [ord('f')],
    "search_input":                     [ord('/')],
    "settings_input":                   [ord('`')],
    "settings_options":                 [ord('~')],
    "continue_search_forward":          [ord('n')],
    "continue_search_backward":         [ord('N')],
    "cancel":                           [27], # Escape key
    "opts_input":                       [ord('e')],
    "opts_select":                      [ord('o')],
    "mode_next":                        [9], # Tab key
    "mode_prev":                        [353], # Shift+Tab key
    "pipe_input":                       [ord('|')],
    "reset_opts":                       [ord('\\')],
    "col_select":                       [ord('0'), ord('1'), ord('2'), ord('3'), ord('4'), ord('5'), ord('6'), ord('7'), ord('8'), ord('9')],
    "col_select_next":                             [ord('>')], 
    "col_select_prev":                             [ord('<')], 
    "col_hide":                         [ord('!'), ord('@'), ord('#'), ord('$'), ord('%'), ord('^'), ord('&'), ord('*'), ord('('), ord(')')],
    "edit":                             [ord('e')],
    "edit_picker":                      [ord('E')],
    "edit_ipython":                     [5], # Ctrl+e
    "copy":                             [ord('y')],
    "paste":                             [ord('p')],
    "save":                             [19, ord('D')],  # Ctrl+s
    "load":                             [15], # Ctrl+o
    "open":                             [ord('O')], 
    "toggle_footer":                    [ord('_')], 
    "notification_toggle":              [ord('z')],
    "redo":                             [ord('.')],
    "undo":                             [26], # Ctrl+z
    "scroll_right":                     [ord('l'), curses.KEY_RIGHT],
    "scroll_right_25":                     [keycodes.META_l],
    "scroll_left_25":                     [keycodes.META_h],
    "scroll_left":                      [ord('h'), curses.KEY_LEFT],
    "scroll_far_right":                 [ord('L')],
    "scroll_far_left":                  [ord('H')],
    # "add_column_before":                       [ord('+')],
    "add_column_after":                       [ord('+')],
    # "add_row_before":                          [ord('=')],
    "add_row_after":                          [ord('=')],
    "info":                             [ord('i')], 
    "file_next":                             [ord('}')], 
    "file_prev":                             [ord('{')], 
    # "sheet_next":                           [],
    # "sheet_prev":                           [],
    "toggle_right_pane":                             [ord("'")], 
    "cycle_right_pane":                             [ord('"')], 
    "toggle_left_pane":                             [ord(";")], 
    "cycle_left_pane":                             [ord(':')], 
}


help_keys = {
    "exit":                             [ord('q'), ord('h')],
    "full_exit":                        [3], # Ctrl+c
    "cursor_down":                      [ord('j'), curses.KEY_DOWN],
    "cursor_up":                        [ord('k'), curses.KEY_UP],
    "half_page_up":                     [ord('u')],
    "half_page_down":                   [ord('d')],
    "page_up":                          [curses.KEY_PPAGE, 2],
    "page_down":                        [curses.KEY_NPAGE, 6],
    "cursor_bottom":                    [ord('G'), curses.KEY_END],
    "cursor_top":                       [ord('g'), curses.KEY_HOME],
    "five_up":                          [ord('K'), keycodes.META_k],
    "five_down":                        [ord('J'), keycodes.META_j],
    "redraw_screen":                    [12], # Ctrl+l
    "cycle_sort_method":                [ord('s')],
    "cycle_sort_method_reverse":        [ord('S')],
    "cycle_sort_order":                 [ord('t')],
    "increase_column_width":            [ord(']')],
    "decrease_column_width":            [ord('[')],
    "filter_input":                     [ord('f')],
    "search_input":                     [ord('/')],
    "settings_input":                   [ord('`')],
    "settings_options":                 [ord('~')],
    "continue_search_forward":          [ord('n')],
    "continue_search_backward":         [ord('N')],
    "cancel":                           [27], # Escape key
    "col_select":                       [ord('0'), ord('1'), ord('2'), ord('3'), ord('4'), ord('5'), ord('6'), ord('7'), ord('8'), ord('9')],
    "col_select_next":                             [ord('>')], 
    "col_select_prev":                             [ord('<')], 
    "col_hide":                         [ord('!'), ord('@'), ord('#'), ord('$'), ord('%'), ord('^'), ord('&'), ord('*'), ord('('), ord(')')],
    "toggle_footer":                    [ord('_')], 
}



notification_keys = {
    "exit":                             [ord('q'), ord('h'), curses.KEY_ENTER, ord('\n'), ord(' '), 27, 13],
    "full_exit":                        [3], # Ctrl+c
    "cursor_down":                      [ord('j'), curses.KEY_DOWN],
    "cursor_up":                        [ord('k'), curses.KEY_UP],
    "half_page_up":                     [ord('u')],
    "half_page_down":                   [ord('d')],
    "page_up":                          [curses.KEY_PPAGE, 2],
    "page_down":                        [curses.KEY_NPAGE, 6],
    "cursor_bottom":                    [ord('G'), curses.KEY_END],
    "cursor_top":                       [ord('g'), curses.KEY_HOME],
    "five_up":                          [ord('K')],
    "five_down":                        [ord('J')],
    "redraw_screen":                    [12], # Ctrl-l
    "refresh":                          [curses.KEY_F5, curses.KEY_RESIZE],
}


menu_keys = {
    "help":                             [ord('?')],
    "exit":                             [ord('q'), ord('h')],
    "full_exit":                        [3], # Ctrl+c
    "cursor_down":                      [ord('j'), curses.KEY_DOWN],
    "cursor_up":                        [ord('k'), curses.KEY_UP],
    "half_page_up":                     [ord('u')],
    "half_page_down":                   [ord('d')],
    "page_up":                          [curses.KEY_PPAGE, 2],
    "page_down":                        [curses.KEY_NPAGE, 6],
    "cursor_bottom":                    [ord('G'), curses.KEY_END],
    "cursor_top":                       [ord('g'), curses.KEY_HOME],
    "five_up":                          [ord('K')],
    "five_down":                        [ord('J')],
    "enter":                            [ord('\n'), curses.KEY_ENTER, ord('l'), 13],
    "redraw_screen":                    [12], # Ctrl-l
    "filter_input":                     [ord('f')],
    "search_input":                     [ord('/')],
    "continue_search_forward":          [ord('n'), ord('i')],
    "continue_search_backward":         [ord('N'), ord('I')],
    "cancel":                           [27], # Escape key
    "opts_input":                       [ord(':')],
    "mode_next":                        [9], # Tab key
    "mode_prev":                        [353], # Shift+Tab key
}


options_keys = {
    "exit":                             [ord('q'), ord('h')],
    "full_exit":                        [3], # Ctrl+c
    "cursor_down":                      [ord('j'), curses.KEY_DOWN],
    "cursor_up":                        [ord('k'), curses.KEY_UP],
    "half_page_up":                     [ord('u')],
    "half_page_down":                   [ord('d')],
    "page_up":                          [curses.KEY_PPAGE, 2],
    "page_down":                        [curses.KEY_NPAGE, 6],
    "cursor_bottom":                    [ord('G'), curses.KEY_END],
    "cursor_top":                       [ord('g'), curses.KEY_HOME],
    "five_up":                          [ord('K')],
    "five_down":                        [ord('J')],
    "toggle_select":                    [ord(' ')],
    "select_all":                       [ord('m'), 1], # Ctrl-a
    "select_none":                      [ord('M'), 18],   # Ctrl-r
    "visual_selection_toggle":          [ord('v')],
    "visual_deselection_toggle":        [ord('V')],
    "enter":                            [ord('\n'), curses.KEY_ENTER, ord('l'), 13],
    "redraw_screen":                    [12], # Ctrl-l
    "cycle_sort_method":                [ord('s')],
    "cycle_sort_method_reverse":        [ord('S')],
    "cycle_sort_order":                 [ord('t')],
    "filter_input":                     [ord('f')],
    "search_input":                     [ord('/')],
    "settings_input":                   [ord('`')],
    "continue_search_forward":          [ord('n'), ord('i')],
    "continue_search_backward":         [ord('N'), ord('I')],
    "cancel":                           [27], # Escape key
    "col_select":                       [ord('0'), ord('1'), ord('2'), ord('3'), ord('4'), ord('5'), ord('6'), ord('7'), ord('8'), ord('9')],
    "refresh":                          [curses.KEY_F5, curses.KEY_RESIZE],
}



edit_menu_keys = {
    "exit":                             [ord('q')],
    "full_exit":                        [3], # Ctrl+c
    "cursor_down":                      [ord('j'), curses.KEY_DOWN],
    "cursor_up":                        [ord('k'), curses.KEY_UP],
    "half_page_up":                     [ord('u')],
    "half_page_down":                   [ord('d')],
    "page_up":                          [curses.KEY_PPAGE, 2], # Ctrl+b
    "page_down":                        [curses.KEY_NPAGE, 6], # Ctrl+f
    "cursor_bottom":                    [ord('G'), curses.KEY_END],
    "cursor_top":                       [ord('g'), curses.KEY_HOME],
    "five_up":                          [ord('K')],
    "five_down":                        [ord('J')],
    "enter":                            [ord('\n'), curses.KEY_ENTER, 13],
    "redraw_screen":                    [12], # Ctrl-l
    "cycle_sort_method":                [ord('s')],
    "cycle_sort_method_reverse":        [ord('S')],
    "cycle_sort_order":                 [ord('t')],
    "increase_column_width":            [ord(']')],
    "decrease_column_width":            [ord('[')],
    "filter_input":                     [ord('f')],
    "search_input":                     [ord('/')],
    "settings_input":                   [ord('`')],
    "settings_options":                 [ord('~')],
    "continue_search_forward":          [ord('n'), ord('i')],
    "continue_search_backward":         [ord('N'), ord('I')],
    "cancel":                           [27], # Escape key
    "opts_input":                       [ord(':')],
    "opts_select":                      [ord('o')],
    "mode_next":                        [9], # Tab key
    "mode_prev":                        [353], # Shift+Tab key
    "pipe_input":                       [ord('|')],
    "reset_opts":                       [ord('\\')],
    "col_select":                       [ord('0'), ord('1'), ord('2'), ord('3'), ord('4'), ord('5'), ord('6'), ord('7'), ord('8'), ord('9')],
    "col_select_next":                             [ord('>'), ord('l')], 
    "col_select_prev":                             [ord('<'), ord('h')], 
    "col_hide":                         [ord('!'), ord('@'), ord('#'), ord('$'), ord('%'), ord('^'), ord('&'), ord('*'), ord('('), ord(')')],
    "edit":                             [ord('e')],
    "edit_picker":                      [ord('E')],
    "edit_ipython":                     [5], # Ctrl+e
    "copy":                             [ord('y')],
    "save":                             [19, ord('D')],  # Ctrl+s
    "load":                             [ord('L'), 15], # Ctrl+o
    "open":                             [ord('O')], 
    "toggle_footer":                    [ord('_')], 
}



from listpick.utils import keycodes
import os, tty, select, curses
import termios

def open_tty():
    """ Return a file descriptor for the tty that we are opening"""
    tty_fd = os.open('/dev/tty', os.O_RDONLY)
    old_terminal_settings = termios.tcgetattr(tty_fd)
    tty.setraw(tty_fd)
    return tty_fd, old_terminal_settings

def restore_terminal_settings(tty_fd, old_settings):
    """ Restore the terminal to its previous state """
    termios.tcsetattr(tty_fd, termios.TCSADRAIN, old_settings)

def get_char(tty_fd, timeout: float = 0.2, secondary: bool = False) -> int:
    """ Get character from a tty_fd with a timeout. """
    rlist, _, _ = select.select([tty_fd], [], [], timeout)
    if rlist:
        # key = ord(tty_fd.read(1))
        key = ord(os.read(tty_fd, 1))
        if not secondary:
            if key == 27:
                key2 = get_char(tty_fd, timeout=0.01, secondary=True)
                key3 = get_char(tty_fd, timeout=0.01, secondary=True)
                key4 = get_char(tty_fd, timeout=0.01, secondary=True)
                key5 = get_char(tty_fd, timeout=0.01, secondary=True)
                key6 = get_char(tty_fd, timeout=0.01, secondary=True)

                keys = [key2, key3, key4, key5, key6]

                key_str = "".join([chr(k) for k in keys if k != -1])

                ## Arrow Keys
                if key2 == ord('O') and key3 == ord('B'):
                    key = curses.KEY_DOWN
                elif key2 == ord('O') and key3 == ord('A'):
                    key = curses.KEY_UP
                elif key2 == ord('O') and key3 == ord('D'):
                    key = curses.KEY_LEFT
                elif key2 == ord('O') and key3 == ord('C'):
                    key = curses.KEY_RIGHT

                ## Shift+ Tab
                elif key2 == ord('[') and key3 == ord('Z'):
                    key = 353

                ## Home, End, Pgup, Pgdn
                elif key2 == ord('O') and key3 == ord('F'):
                    key = curses.KEY_END
                elif key2 == ord('O') and key3 == ord('H'):
                    key = curses.KEY_HOME
                elif key2 == ord('[') and key3 == ord('6') and key4 == ord("~"):
                    key = curses.KEY_NPAGE
                elif key2 == ord('[') and key3 == ord('5') and key4 == ord("~"):
                    key = curses.KEY_PPAGE


                # Delete key
                elif key_str == "[3~":    ## Delete
                    key = curses.KEY_DC
                elif key_str == "[3;2~":                                            ## Shift+Delete
                    key = 383

                ## Function Keys
                elif key2 == ord('O') and key3 == ord('P'):
                    key = curses.KEY_F1
                elif key_str == "OQ":
                    key = curses.KEY_F2
                elif key_str == "OR":
                    key = curses.KEY_F3
                elif key_str == "OS":
                    key = curses.KEY_F4
                elif key_str == "[15~":
                    key = curses.KEY_F5
                elif key_str == "[17~":
                    key = curses.KEY_F6
                elif key_str == "[17~":
                    key = curses.KEY_F7
                elif key_str == "[19~":
                    key = curses.KEY_F8
                elif key_str == "[20~":
                    key = curses.KEY_F9
                elif key_str == "[21~":
                    key = curses.KEY_F10
                elif key_str == "[23~":
                    key = curses.KEY_F11
                elif key_str == "[24~":
                    key = curses.KEY_F12

                ## Alt+KEY
                elif key2 >= ord('a') and key2 <= ord('z') and key3 == -1:      ## Alt+[a-zA-Z]
                    key = keycodes.META_a + (key2 - ord('a'))
                elif key2 >= ord('A') and key2 <= ord('Z') and key3 == -1:      ## Alt+[a-zA-Z]
                    key = keycodes.META_A + (key2 - ord('A'))
                elif key2 == ord('0') and key3 == -1:                           ## Alt+0
                    key = keycodes.META_0
                elif key2 >= ord('1') and key2 <= ord('9') and key3 == -1:      ## Alt+1-9
                    key = keycodes.META_1 + (key2 - ord('1'))
                elif key2 == 127:                                               ## Alt+BS
                    key = keycodes.META_BS

                # If it is an unknown key with an escape sequence then return -1.
                elif key2 != -1:
                    key = -1
        

    else:
        key = -1
    return key



#!/bin/python
# -*- coding: utf-8 -*-
"""
generate_data_multithreaded.py
Generate data for listpick Picker.


1. Read toml file.
2. Set environment variables.
3. Get files from first command.
4. Create items with "..." for cells to be filled.
5. Create a priority queue which determines which cells are to be filled first.
6. Create a queue updater which increases the priorty of cells which are on screen which does so each second.
7. Create threads to start generating data for cells.

Author: GrimAndGreedy
License: MIT
"""

import subprocess
import os
from typing import Tuple, Callable
import toml
import logging
import threading
from queue import PriorityQueue
import time
import re
import shlex

logger = logging.getLogger('picker_log')

def generate_columns_worker(
    funcs: list,
    files: list,
    items: list[list[str]],
    getting_data: threading.Event,
    task_queue: PriorityQueue,
    completed_cells: set,
) -> None:
    """ Get a task from the priorty queue and fill the data for that cell."""
    while task_queue.qsize() > 0:
        _, (i, j) = task_queue.get()

        if (i, j) in completed_cells:
            task_queue.task_done()
            continue
            
        generate_cell(
            func=funcs[j],
            file=files[i],
            items=items,
            row=i,
            col=j+1,
        )
        completed_cells.add((i, j))
        task_queue.task_done()
    getting_data.set()

def generate_cell(func: Callable, file: str, items: list[list[str]], row: int, col: int) -> None:
    """
    Takes a function, file and a file and then sets items[row][col] to the result.
    """
    
    items[row][col] = func(file).strip()

def update_queue(task_queue: PriorityQueue, visible_rows_indices: list[int], rows: int, cols: int):
    """ Increase the priority of getting the data for the cells that are currently visible. """
    while task_queue.qsize() > 0:
        time.sleep(0.1)
        for row in visible_rows_indices:
            for col in range(cols):
                if 0 <= row < rows:
                    task_queue.put((1, (row, col)))

        # Delay
        time.sleep(0.9)

def command_to_func(command: str) -> Callable:
    """
    Convert a command string to a function that will run the command.

    E.g.,
        mediainfo {} | grep -i format
        mediainfo {} | grep -i format | head -n 1 | awk '{{print $3}}'
    """
    logger.info("function: command_to_func (generate_data.py)")
    
    func = lambda arg: subprocess.run(replace_braces(command, arg), shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE).stdout.decode("utf-8").strip()
    return func

def load_environment(envs:dict):
    """
    Load environment variables from an envs dict.
    """
    logger.info("function: load_environment (generate_data.py)")

    if "cwd" in envs:
        os.chdir(os.path.expandvars(os.path.expanduser(envs["cwd"])))

def replace_braces(text, s):
    text = re.sub(r'\{\{(.*?)\}\}', r'@@\1@@', text)
    text = re.sub(r'\{\}', shlex.quote(s), text)
    text = re.sub(r'@@(.*?)@@', r'{{\1}}', text)
    return text

def read_toml(file_path) -> Tuple[dict, list, list]:
    """
    Read toml file and return the environment, commands and header sections.
    """
    logger.info("function: read_toml (generate_data.py)")
    with open(file_path, 'r') as file:
        config = toml.load(file)

    environment = config['environment'] if 'environment' in config else {}
    data = config['data'] if 'data' in config else {}
    commands = [command.strip() for command in data['commands']] if 'commands' in data else []
    header = [header for header in data['header']]  if 'header' in data else []
    return environment, commands, header
    
    
def generate_picker_data_from_file(
    file_path: str,
    items,
    header,
    visible_rows_indices,
    getting_data
) -> None:
    """
    Generate data for Picker based upon the toml file commands.
    """
    logger.info("function: generate_picker_data (generate_data.py)")


    environment, lines, hdr = read_toml(file_path)

    # Load any environment variables from the toml file
    if environment:
        load_environment(environment)


    # Get list of files to be displayed in the first column.
    get_files_command = lines[0]
    files = subprocess.run(get_files_command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE).stdout.decode("utf-8").strip().split("\n")
    files = [file.strip() for file in files if files]
    
    commands_list = [line.strip() for line in lines[1:]]
    command_funcs = [command_to_func(command) for command in commands_list]

    generate_picker_data(
        files = files,
        column_functions = command_funcs,
        data_header = hdr,
        items = items,
        picker_header = header,
        visible_rows_indices = visible_rows_indices,
        getting_data = getting_data,
    )

def generate_picker_data(
    files: list[str],
    column_functions: list[Callable],
    data_header,
    items,
    picker_header,
    visible_rows_indices,
    getting_data
) -> None:
    """
    Generate data from a list of files and a list of column functions which will be used to 
        generate subsequent columns.

    This function is performed asynchronously with os.cpu_count() threads.

    data_header: header list to be set for the picker
    picker_header: the picker header will be passed in so that it can be set for the class

    """
    logger.info("function: generate_picker_data (generate_data.py)")

    items.clear()
    items.extend([[file] + ["..." for _ in column_functions] for file in files])
    # items[:] = [[file] + ["..." for _ in column_functions] for file in files]
    picker_header[:] = data_header


    task_queue = PriorityQueue()
    for i in range(len(files)):
        for j in range(len(column_functions)):
            task_queue.put((10, (i, j)))

    num_workers = os.cpu_count()
    if num_workers in [None, -1]: num_workers = 4
    if num_workers == None or num_workers < 1: num_workers = 1
    completed_cells = set()

    for _ in range(num_workers):
        gen_items_thread = threading.Thread(
            target=generate_columns_worker,
            args=(column_functions, files, items, getting_data, task_queue, completed_cells),
        )
        gen_items_thread.daemon = True
        gen_items_thread.start()

    update_queue_thread = threading.Thread(
        target=update_queue,
        args=(task_queue, visible_rows_indices, len(files), len(column_functions)),
    )
    update_queue_thread.daemon = True
    update_queue_thread.start()



#!/bin/python
# -*- coding: utf-8 -*-
"""
table_to_lists.py

Author: GrimAndGreedy
License: MIT
"""

import sys
import csv
import json
from io import StringIO
import argparse
from typing import Tuple, Iterable, Optional
import dill as pickle
import os
import logging

logger = logging.getLogger('picker_log')

def read_file_content(file_path: str) -> str:
    """ Read lines from file. """
    logger.info("function: read_file_content (table_to_list_of_lists.py)")
    with open(file_path, 'r') as file:
        return file.read()

def strip_whitespace(item: Iterable) -> Iterable:
    """ Strip whitespace from string or from list of strings. """
    logger.info("function: strip_whitespace (table_to_list_of_lists.py)")
    if isinstance(item, list):
        return [strip_whitespace(sub_item) for sub_item in item]
    elif isinstance(item, str):
        return item.strip()
    else:
        return item


def xlsx_to_list(file_name: str, sheet_number:int = 0, extract_formulae: bool = False, first_row_is_header: bool = True):
    import pandas as pd
    from openpyxl import load_workbook
    # wb = load_workbook(filename=input_arg, read_only=True)
    # values or formulae
    if not os.path.exists(file_name):
        return [], [], []
    wb = load_workbook(filename=file_name, read_only=True, data_only=not extract_formulae)
    
    if not isinstance(sheet_number, int): sheet_number = 0
    sheet_number = max(0, min(sheet_number, len(wb.sheetnames)-1))
    ws = wb.worksheets[sheet_number]
    
    # Read data and formulas from the sheet
    table_data = []
    # table_data = [[cell for cell in row] for row in ws.iter_rows(min_row=1, values_only=False)]
    # table_data = [[cell.value for cell in row] for row in ws.iter_rows(min_row=1, values_only=False)]
    table_data = [[cell if cell != None else "" for cell in row] for row in ws.iter_rows(min_row=1, values_only=True)]
    header = []
    # header = [cell for cell in list(ws.iter_rows(values_only=True))[0]]  # Assuming the first row is the header
    if first_row_is_header and len(table_data) > 1:
        header = table_data[0]
        table_data = table_data[1:]
    else:
        header = []
    #
    # for row in ws.iter_rows(min_row=2, values_only=True):  # Skip the header row
    #     row_data = []
    #     for cell in row:
    #         if isinstance(cell, str) and '=' in cell:  # Check if it's a formula
    #             row_data.append(cell)
    #         else:
    #             row_data.append(str(cell))
    #     table_data.append(row_data)
    
    return table_data, header, wb.sheetnames

def ods_to_list(filename: str, sheet_number: int = 0, extract_formulas: bool = False, first_row_is_header: bool = True):
    from odf.opendocument import load
    from odf import table, text

    from odf.namespaces import TABLENS
    # Load the ODS file
    doc = load(filename)

    sheets = doc.spreadsheet.getElementsByType(table.Table)
    sheet_names = [s.attributes.get((TABLENS, 'name')) for s in sheets]


    # Get the sheet by index
    sheet = doc.spreadsheet.getElementsByType(table.Table)[sheet_number]

    data = []
    for row in sheet.getElementsByType(table.TableRow):
        row_data = []
        for cell in row.getElementsByType(table.TableCell):
            if extract_formulas:
                formula = cell.attributes.get((TABLENS, 'formula'))
                if formula is not None:
                    row_data.append(formula)
                    continue  # Skip extracting value if formula found
            
            # Extract value (as text) from <text:p> elements
            cell_text = ""
            for p in cell.getElementsByType(text.P):
                cell_text += str(p.firstChild) if p.firstChild is not None else ""
            row_data.append(cell_text)
        data.append(row_data)
    if first_row_is_header and len(data) > 1:
        header = data[0]
        data = data[1:]
    else:
        header = []

    return data, header, sheet_names

def ods_to_list_old(file_name: str, sheet_number:int = 0, extract_formulae: bool = False, first_row_is_header: bool = True):
    try:
        import pandas as pd
        ef = pd.ExcelFile(file_name)
        sheets = ef.sheet_names
        sheet_number = max(0, min(sheet_number, len(sheets)-1))
        df = pd.read_excel(file_name, engine='odf', sheet_name=sheet_number)
        # if sheet_number < len(sheets):
        #     df = pd.read_excel(input_arg, engine='odf', sheet_name=sheet_number)
        # else:
        #     df = pd.read_excel(input_arg, engine='odf')
        table_data = df.values.tolist()
        table_data = [[x if not pd.isna(x) else "" for x in row] for row in table_data]
        
        try:
            header = list(df.columns)
        except:
            header = []

        if not first_row_is_header and header:
            table_data = [header] + table_data
            header = []

        return table_data, header, sheets
    except Exception as e:
        print(f"Error loading ODS file: {e}")
        return [], [], []


def table_to_list(
    input_arg: str,
    delimiter:str='\t',
    file_type:Optional[str]=None,
    sheet_number:int = 0,
    first_row_is_header:bool = True,

) -> Tuple[list[list[str]], list[str], list[str]]:
    """ 
    Convert data string to list. The input_arg
    Currently accepts: csv, tsv, json, xlsx, ods


    input_arg: filename


    returns:
        items: list[list[str]]
        header: list[str]
        sheets: list[str]
    
    """
    logger.info("function: table_to_list (table_to_list_of_lists.py)")
    table_data = []

    def parse_csv_like2(data:str, delimiter:str) -> list[list[str]]:
        """ Convert value-separated data (e.g., CSV or TSV) to list of lists. """
        logger.info("function: parse_csv_like (table_to_list_of_lists.py)")

        try:
            # reader = csv.reader(StringIO(data), delimiter=delimiter)
            reader = csv.reader(StringIO(data), dialect='unix')
            return [row for row in reader]
        except Exception as e:
            print(f"Error reading CSV-like input: {e}")
            return []
    def parse_csv_like(data:str, delimiter: str=" "):
        import re
        def split_columns(line):
            # Define the regex pattern to match quoted strings and split by whitespace
            # pattern = r"(?:'[^']*'|[^'\s]+)"
            pattern = r"(?:\"[^\"]*\"|'[^']*'|[^'\s]+)"
            
            # Find all matches using the defined pattern
            columns = re.findall(pattern, line)
    
            return columns

        lines = data.strip().split('\n')
        result = []
        
        for line in lines:
            result.append(split_columns(line))
        
        return result

    def csv_string_to_list(csv_string:str, first_row_is_header: bool = True) -> list[list[str]]:
        """ Convert csv string to list of lists using csv.reader. """
        logger.info("function: csv_string_to_list (table_to_list_of_lists.py)")
        f = StringIO(csv_string)
        reader = csv.reader(f, skipinitialspace=True)
        table_data = [row for row in reader]
        if first_row_is_header and len(table_data) > 1:
            header = table_data[0]
            table_data = table_data[1:]
        else:
            header = []
        return table_data, header
            

    if input_arg == '--stdin':
        input_data = sys.stdin.read()
    elif input_arg == '--stdin2':
        input_count = int(sys.stdin.readline())
        input_data = "\n".join([sys.stdin.readline() for i in range(input_count)])
        sys.stdin.flush()
        # sys.stdin.close()
        # sys.stdin = open('/dev/tty', 'r')
    elif file_type == 'csv' or delimiter in [',']:
        try:
            if input_arg == '--stdin':
                input_data = sys.stdin.read()
            elif input_arg == '--stdin2':
                input_count = int(sys.stdin.readline())
                input_data = "\n".join([sys.stdin.readline().strip() for i in range(input_count)])
            else:
                input_data = read_file_content(input_arg)

            table_data, header = csv_string_to_list(input_data, first_row_is_header)
            table_data = strip_whitespace(table_data)
            header = strip_whitespace([header])[0]
            # table_data = parse_csv_like(input_data, ",")
            return table_data, header, []
        except Exception as e:
            print(f"Error reading CSV/TSV input: {e}")
            return [], [], []

    elif file_type == 'tsv':
        try:
            if input_arg == '--stdin':
                input_data = sys.stdin.read()
            elif input_arg == '--stdin2':
                input_count = int(sys.stdin.readline())
                input_data = "\n".join([sys.stdin.readline().strip() for i in range(input_count)])
            else:
                input_data = read_file_content(input_arg)
            
            # Adjust delimiter for TSV or CSV
            if file_type == 'tsv' or delimiter == '\t':
                delimiter = '\t'
            else:
                delimiter = ','
            
            table_data = parse_csv_like(input_data, delimiter)
            table_data = strip_whitespace(table_data)
            return table_data, [], []
        except Exception as e:
            print(f"Error reading CSV/TSV input: {e}")
            return [], [], []

    elif file_type == 'json':
        try:
            if input_arg == '--stdin':
                input_data = sys.stdin.read()
            elif input_arg == '--stdin2':
                input_count = int(sys.stdin.readline())
                input_data = "\n".join([sys.stdin.readline() for i in range(input_count)])
            else:
                input_data = read_file_content(input_arg)
            
            table_data = json.loads(input_data)
            return table_data, [], []
        except json.JSONDecodeError as e:
            print(f"Error decoding JSON input: {e}")
            return [], [], []
        except FileNotFoundError as e:
            print(f"File not found: {e}")
            return [], [], []

    elif file_type == 'xlsx':
        extract_formulae = False
        return xlsx_to_list(input_arg, sheet_number, extract_formulae, first_row_is_header)

    elif file_type == 'ods':
        extract_formulae = False
        return ods_to_list(input_arg, sheet_number, extract_formulae, first_row_is_header)
    elif file_type == 'pkl':
        with open(os.path.expandvars(os.path.expanduser(input_arg)), 'rb') as f:
            loaded_data = pickle.load(f)
        items = loaded_data["items"] if "items" in loaded_data else []
        header = loaded_data["header"] if "header" in loaded_data else []
        return items, header, []

    else:
        input_data = read_file_content(input_arg)
    
    table_data = parse_csv_like(input_data, delimiter)
    if first_row_is_header and len(table_data) > 1:
        header = table_data[0]
        table_data = table_data[1:]
    else:
        header = []


    return table_data, header, []

if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Convert table to list of lists.')
    parser.add_argument('-i', dest='file', help='File containing the table to be converted')
    parser.add_argument('--stdin', action='store_true', help='Table passed on stdin')
    parser.add_argument('--stdin2', action='store_true', help='Table passed on stdin')
    parser.add_argument('-d', dest='delimiter', default='\t', help='Delimiter for rows in the table (default: tab)')
    parser.add_argument('-t', dest='file_type', choices=['tsv', 'csv', 'json', 'xlsx', 'ods'], help='Type of file (tsv, csv, json, xlsx, ods)')
    
    args = parser.parse_args()
    
    if args.file:
        input_arg = args.file
    elif args.stdin:
        input_arg = '--stdin'
    elif args.stdin2:
        input_arg = '--stdin2'
    else:
        print("Error: Please provide input file or use --stdin option.")
        sys.exit(1)
    
    table_data = table_to_list(input_arg, args.delimiter, args.file_type)
    # print(table_data)

    len(table_data[0])
    for row in table_data:
        if len(row) != len(table_data[0]):
            print(len(row), row)
